==> diablo-1.05.tgz: diablo/util/dexpire.c <==

/*
 * DEXPIRE.C	- diablo expire.
 *
 *	Runs through the history database looking for articles to
 *	expire.  When we find one, and if ROpt is true, we remove
 *	the file and update the history database (set the
 *	exp field to ((unsigned short)-1)).  If a free-space target
 *	has been given, we check the free space on the spool partition
 *	and extend the expiration and loop until the target is met.
 *
 *	Unlike INN, this program does NOT rewrite the history file.
 *	It simply modifies it in-place in a manner that does not
 *	interfere with nominal diablo server operation.
 *
 * (c)Copyright 1997, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 */

#include "defs.h"
#include <sys/param.h>
#include <sys/mount.h>
#if USE_SYSV_STATFS
#include <sys/statfs.h>
#define f_bavail f_bfree
#endif

int FOpt;
int VOpt = -1;
int TOpt;
int ROpt;
int FreeSpaceTarget;
int LotsOfFreeSpace = 1;

void DoChildRemove(int pfd, int fd);
int NotEnoughFreeSpace(void);

int
main(int ac, char **av)
{
    History hary[16384];
    int n;
    int fd;
    double factor = 1.0;

    {
	FILE *fi;

	if ((fi = xfopen("r", "%s/dexpire.factor", NewsHome)) != NULL) {
	    fscanf(fi, "%lf", &factor);
	    fclose(fi);
	}
    }

    for (n = 1; n < ac; ++n) {
	char *ptr = av[n];
	if (*ptr == '-') {
	    ptr += 2;
	    switch(ptr[-1]) {
	    case 'a':
		factor = strtod((*ptr ? ptr : av[++n]), NULL);
		break;
	    case 'r':
		FreeSpaceTarget = strtol((*ptr ? ptr : av[++n]), NULL, 0);	/* in MB */
		ROpt = 1;
		break;
	    case 'q':
		VOpt = 0;
		break;
	    case 't':
		TOpt = 1;
		FOpt = 1;
		VOpt = 1;
		break;
	    case 'v':
		VOpt = 1;
		break;
	    case 'f':
		FOpt = 1;
		break;
	    default:
		printf("Illegal option: %s\n", ptr - 2);
		exit(1);
	    }
	} else {
	    printf("Illegal argument: %s\n", ptr);
	    exit(1);
	}
    }

    if (ac == 1) {
	printf("dexpire [-f] [-v] [-t] [-a addhrs] [-r0] [-r freespacetarget]\n");
	exit(0);
    }

    if (chdir(SpoolHome) != 0) {
	perror("chdir");
	exit(1);
    }

    fprintf(stderr, "dexpire: starting factor %6.2f\n", factor);

    do {
	int hcount = 0;
	int hold = 0;
	int hexp = 0;
	int fds[2];
	int hsize = 1024 * 1024;
	pid_t pid;

	/*
	 * Create a pipe between parent and child.
	 * The parent will remove the files while the
	 * child updates the dhistory file.
	 */

	if (pipe(fds) != 0) {
	    perror("pipe");
	    exit(1);
	}

	fflush(stdout);
	fflush(stderr);

	/*
	 * Fork.  Note that we open the dhistory file
	 * AFTER we fork, because otherwise the seek
	 * position for the fd will be shared across processes.
	 */

	pid = fork();
	if (pid < 0) {
	    perror("fork");
	    exit(1);
	}

	fd = xopen(O_RDWR, 0644, "%s/dhistory", NewsHome);

	{
	    FILE *fi = xfopen("r", "%s/dhistory.param", NewsHome);
	    if (fi) {
		char buf[256];

		hsize = 0;
		while (fgets(buf, sizeof(buf), fi) != NULL) {
		    if (buf[0] == 'H')
			hsize = strtol(buf + 1, NULL, 0);
		}
		if (hsize == 0) {
		    fprintf(stderr, "dhistory parameter file error\n");
		    exit(1);
		}
		fclose(fi);
		if (pid == 0)
		    printf("hash size: %d\n", hsize);
	    }
	}

	if (fd < 0) {
	    perror("open(dhistory)");
	    exit(1);
	}

	if (pid == 0) {
	    close(fds[1]);
	    DoChildRemove(fds[0], fd);
	    exit(0);
	}

	close(fds[0]);

	if (TOpt)
	    lseek(fd, 0L, 2);
	else
	    lseek(fd, hsize * sizeof(long), 0);

	for (;;) {
	    unsigned long gmt = time(NULL) / 60;

	    while ((n = read(fd, hary, sizeof(hary))) > 0) {
		int i;
		long npos = lseek(fd, 0L, 1);
		long opos = npos - n;

		n /= sizeof(hary[0]);

		for (i = 0; i < n; ++i) {
		    History *h = &hary[i];
		    double dtime;

		    /*
		     * calculate dtime in hours.  A positive number means we have not
		     * yet expired.  A negative number means we have.
		     */
		    dtime = ((double)h->gmt + factor * (double)(h->exp * 60) - (double)gmt) / 60.0;

		    ++hcount;

		    if (h->exp != (unsigned short)-1) {
			if (dtime < 0.0) {
			    char path[256];
			    int rv;

			    ++hexp;

			    /*
			     * Figure out the filepath
			     */

			    sprintf(path, "D.%08lx/%08lx.%08lx.%04x",
				h->gmt - h->gmt % 10,
				h->hv.h1,
				h->hv.h2,
				h->iter
			    );
			    if (VOpt) {
				printf("%s exp=%-3d rgmt=%-6.2f", 
				    path, 
				    h->exp,
				    dtime
				);
			    }

			    /*
			     * Delete the file and update the exp field in the history
			     * if appropriate.
			     *
			     * TODO: use chdir to cache the directory, then only remove
			     * files in a manner simliar to INND's fastrm.
			     */

			    if (ROpt && ((rv=remove(path)) == 0 || errno == ENOENT || errno == ENOTDIR)) {
				long r = opos + i * sizeof(History) + offsetof(History, exp);

				write(fds[1], &r, sizeof(long));

				if (VOpt) {
				    if (rv == 0)
					printf("\t(removed)");
				    else
					printf("\t(%s)", strerror(errno));
				}
			    }
			    if (VOpt)
				printf("\n");
			} else {
			    /*
			     * this is just a verbose dump of the history file
			     */

			    if (VOpt > 0) {
				printf(
				    "D.%08lx/%08lx.%08lx.%04x in %4.2f hours exp=%d\n",
				    h->gmt - h->gmt % 10,
				    h->hv.h1,
				    h->hv.h2,
				    h->iter,
				    dtime,
				    (int)h->exp
				);
			    }
			}
		    } else {
			++hold;
		    }
		}
	    }
	    close(fd);
	    if (FOpt == 0)
		break;
	    if (TOpt)
		VOpt = TOpt;
	    sleep(1);
	}
	close(fds[1]);
	waitpid(pid, NULL, 0);

	sleep(10);

	fprintf(stderr, "Scanned f=%-6.2f count=%-6d oldexp=%-6d expired=%-6d\n",
	    factor,
	    hcount,
	    hold,
	    hexp
	);
	factor -= 0.05;
    } while (factor >= 0.0 && NotEnoughFreeSpace());

    if (LotsOfFreeSpace)
	factor += 0.05 + 0.05;

    /*
     * print the new factor.  If there was enough free space the
     * first time, print a new factor + 0.05
     */
    {
	char path1[256];
	char path2[256];
	FILE *fo;

	sprintf(path1, "%s/dexpire.factor.%d", NewsHome, (int)getpid());
	sprintf(path2, "%s/dexpire.factor", NewsHome);
	if ((fo = fopen(path1, "w")) != NULL) {
	    fprintf(fo, "%6.2f\n", factor);
	    fflush(fo);
	    if (ferror(fo) == 0)
		rename(path1, path2);
	    else
		remove(path1);
	    fclose(fo);
	}
    }

    fprintf(stderr, "dexpire: ending factor %6.2f\n", factor);

    return(0);
}

/*
 * calculate the free space and adjust the expiration munge
 * accordingly.
 */

int
NotEnoughFreeSpace(void)
{
    struct statfs stmp;
    int avail;
    char path[1024];

    sprintf(path, "%s/.", SpoolHome);

    if (FreeSpaceTarget == 0)
	return(0);
#if USE_SYSV_STATFS
    if (statfs(path, &stmp, sizeof(stmp), 0) != 0) {
#else
    if (statfs(path, &stmp) != 0) {
#endif
	fprintf(stderr, "dexpire: unable to statfs %s\n", path);
	return(0);
    }
    avail = stmp.f_bavail * (stmp.f_bsize / 512) / 2;	/* available space in KB */
    fprintf(stderr, "%4.2f MB free space availble\n", (double)avail / 1024.0);
    if (avail / 1024 < FreeSpaceTarget) {
	LotsOfFreeSpace = 0;
	return(1);
    }
    return(0);

#ifdef NOTDEF
	if (stmp.f_bavail < sfs.f_bavail)
	    sfs.f_bavail = stmp.f_bavail;
	if (stmp.f_ffree < sfs.f_ffree) {
	    sfs.f_ffree = stmp.f_ffree;
	    sfs.f_files = stmp.f_files;
	}
    }
#endif
}

void 
DoChildRemove(int pfd, int fd)
{
    long r;
    unsigned short nexp = (unsigned short)-1;

    while (read(pfd, &r, sizeof(long)) == sizeof(long)) {
       if (lseek(fd, r, 0) == r)
	   write(fd, &nexp, sizeof(nexp));
    }
}

==> diablo-1.06.tgz: diablo/util/dexpire.c <==

/*
 * DEXPIRE.C	- diablo expire.
 *
 *	Runs through the history database looking for articles to
 *	expire.  When we find one, and if ROpt is true, we remove
 *	the file and update the history database (set the
 *	exp field to ((unsigned short)-1)).  If a free-space target
 *	has been given, we check the free space on the spool partition
 *	and extend the expiration and loop until the target is met.
 *
 *	Unlike INN, this program does NOT rewrite the history file.
 *	It simply modifies it in-place in a manner that does not
 *	interfere with nominal diablo server operation.
 *
 * (c)Copyright 1997, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 */

#include "defs.h"
#include <sys/param.h>
#include <sys/mount.h>
#if USE_SYSV_STATFS
#include <sys/statfs.h>
#define f_bavail f_bfree
#endif

int FOpt;
int VOpt = -1;
int TOpt;
int ROpt;
int FreeSpaceTarget;
int LotsOfFreeSpace = 1;
int ParallelOpt = 1;

int SkipOpt = 1;
int IndexOpt = 0;

void DoChildRemove(int pfd, int fd);
int NotEnoughFreeSpace(void);

int
main(int ac, char **av)
{
    History hary[16384];
    int n;
    int fd;
    double factor = 1.0;

    {
	FILE *fi;

	if ((fi = xfopen("r", "%s/dexpire.factor", NewsHome)) != NULL) {
	    fscanf(fi, "%lf", &factor);
	    fclose(fi);
	}
    }

    for (n = 1; n < ac; ++n) {
	char *ptr = av[n];
	if (*ptr == '-') {
	    ptr += 2;
	    switch(ptr[-1]) {
	    case 'a':
		factor = strtod((*ptr ? ptr : av[++n]), NULL);
		break;
	    case 'r':
		FreeSpaceTarget = strtol((*ptr ? ptr : av[++n]), NULL, 0);	/* in MB */
		ROpt = 1;
		break;
	    case 'q':
		VOpt = 0;
		break;
	    case 't':
		TOpt = 1;
		FOpt = 1;
		VOpt = 1;
		break;
	    case 'v':
		VOpt = 1;
		break;
	    case 'f':
		FOpt = 1;
		break;
	    case 'p':
		ParallelOpt = strtol((*ptr ? ptr : av[++n]), NULL, 0);
		break;
	    default:
		printf("Illegal option: %s\n", ptr - 2);
		exit(1);
	    }
	} else {
	    printf("Illegal argument: %s\n", ptr);
	    exit(1);
	}
    }

    if (ac == 1) {
	printf("dexpire [-f] [-v] [-t] [-a addhrs] [-r0] [-r freespacetarget] [-p #tasks]\n");
	exit(0);
    }

    if (chdir(SpoolHome) != 0) {
	perror("chdir");
	exit(1);
    }

    /*
     * If ParallelOpt is > 1, fork as appropriate
     */

    fflush(stdout);
    fflush(stderr);

    if (ParallelOpt > 1) {
	pid_t *pidAry = malloc(sizeof(pid_t) * ParallelOpt);

	SkipOpt = ParallelOpt;
	for (IndexOpt = 0; IndexOpt < SkipOpt; ++IndexOpt) {
	    if ((pidAry[IndexOpt] = fork()) == 0)
		break;
	}
	if (IndexOpt == SkipOpt) {
	    /*
	     * wait for one to exit
	     */
	    while (wait(NULL) < 0)
		;
	    /*
	     * kill *all* of them
	     */
	    while (--SkipOpt >= 0) {
		if (pidAry[SkipOpt] > 0)
		    kill(pidAry[SkipOpt], 9);
	    }
	    exit(0);
	}
    }

    fprintf(stderr, "dexpire: starting factor %6.2f\n", factor);

    do {
	int hcount = 0;
	int hold = 0;
	int hexp = 0;
	long size = 0;
	int fds[2];
	int hsize = 1024 * 1024;
	pid_t pid;

	/*
	 * Create a pipe between parent and child.
	 * The parent will remove the files while the
	 * child updates the dhistory file.
	 */

	if (pipe(fds) != 0) {
	    perror("pipe");
	    exit(1);
	}

	fflush(stdout);
	fflush(stderr);

	/*
	 * Fork.  Note that we open the dhistory file
	 * AFTER we fork, because otherwise the seek
	 * position for the fd will be shared across processes.
	 */

	pid = fork();
	if (pid < 0) {
	    perror("fork");
	    exit(1);
	}

	fd = xopen(O_RDWR, 0644, "%s/dhistory", NewsHome);

	{
	    FILE *fi = xfopen("r", "%s/dhistory.param", NewsHome);
	    if (fi) {
		char buf[256];

		hsize = 0;
		while (fgets(buf, sizeof(buf), fi) != NULL) {
		    if (buf[0] == 'H')
			hsize = strtol(buf + 1, NULL, 0);
		}
		if (hsize == 0) {
		    fprintf(stderr, "dhistory parameter file error\n");
		    exit(1);
		}
		fclose(fi);
		if (pid == 0)
		    printf("hash size: %d\n", hsize);
	    }
	}

	if (fd < 0) {
	    perror("open(dhistory)");
	    exit(1);
	}

	if (pid == 0) {
	    close(fds[1]);
	    DoChildRemove(fds[0], fd);
	    exit(0);
	}

	close(fds[0]);

	if (TOpt) {
	    lseek(fd, 0L, 2);
	} else {
	    struct stat st;

	    lseek(fd, hsize * sizeof(long), 0);
	    if (fstat(fd, &st) == 0)
		size = st.st_size;	/* limit scan */
	}

	for (;;) {
	    unsigned long gmt = time(NULL) / 60;

	    while ((n = read(fd, hary, sizeof(hary))) > 0) {
		int i;
		long npos = lseek(fd, 0L, 1);
		long opos = npos - n;

		/*
		 * we can get stuck in an endless loop if
		 * diablo is appending to the file while
		 * we are reading from it.  It actually
		 * happens!  So we limit each scan run
		 * to the size of the dhistory file as of
		 * the start of the run.
		 */

		if (size && opos >= size)
		    break;

		n /= sizeof(hary[0]);

		for (i = IndexOpt; i < n; i += SkipOpt) {
		    History *h = &hary[i];
		    double dtime;

		    /*
		     * calculate dtime in hours.  A positive number means we have not
		     * yet expired.  A negative number means we have.
		     */
		    dtime = ((double)h->gmt + factor * (double)(h->exp * 60) - (double)gmt) / 60.0;

		    ++hcount;

		    if (h->exp != (unsigned short)-1) {
			if (dtime < 0.0) {
			    char path[256];
			    int rv;

			    ++hexp;

			    /*
			     * Figure out the filepath
			     */

			    sprintf(path, "D.%08lx/%08lx.%08lx.%04x",
				h->gmt - h->gmt % 10,
				h->hv.h1,
				h->hv.h2,
				h->iter
			    );
			    if (VOpt) {
				printf("%s exp=%-3d rgmt=%-6.2f", 
				    path, 
				    h->exp,
				    dtime
				);
			    }

			    /*
			     * Delete the file and update the exp field in the history
			     * if appropriate.
			     *
			     * TODO: use chdir to cache the directory, then only remove
			     * files in a manner simliar to INND's fastrm.
			     */

			    if (ROpt && ((rv=remove(path)) == 0 || errno == ENOENT || errno == ENOTDIR)) {
				long r = opos + i * sizeof(History) + offsetof(History, exp);

				write(fds[1], &r, sizeof(long));

				if (VOpt) {
				    if (rv == 0)
					printf("\t(removed)");
				    else
					printf("\t(%s)", strerror(errno));
				}
			    }
			    if (VOpt)
				printf("\n");
			} else {
			    /*
			     * this is just a verbose dump of the history file
			     */

			    if (VOpt > 0) {
				printf(
				    "D.%08lx/%08lx.%08lx.%04x in %4.2f hours exp=%d\n",
				    h->gmt - h->gmt % 10,
				    h->hv.h1,
				    h->hv.h2,
				    h->iter,
				    dtime,
				    (int)h->exp
				);
			    }
			}
		    } else {
			++hold;
		    }
		}
	    }
	    if (FOpt == 0)
		break;
	    if (TOpt)
		VOpt = TOpt;
	    sleep(1);
	}
	close(fd);
	close(fds[1]);
	waitpid(pid, NULL, 0);

	sleep(10);

	fprintf(stderr, "Scanned f=%-6.2f count=%-6d oldexp=%-6d expired=%-6d\n",
	    factor,
	    hcount,
	    hold,
	    hexp
	);
	factor -= 0.05;
    } while (factor >= 0.0 && NotEnoughFreeSpace());

    if (LotsOfFreeSpace)
	factor += 0.05 + 0.05;

    /*
     * print the new factor.  If there was enough free space the
     * first time, print a new factor + 0.05
     */
    {
	char path1[256];
	char path2[256];
	FILE *fo;

	sprintf(path1, "%s/dexpire.factor.%d", NewsHome, (int)getpid());
	sprintf(path2, "%s/dexpire.factor", NewsHome);
	if ((fo = fopen(path1, "w")) != NULL) {
	    fprintf(fo, "%6.2f\n", factor);
	    fflush(fo);
	    if (ferror(fo) == 0)
		rename(path1, path2);
	    else
		remove(path1);
	    fclose(fo);
	}
    }

    fprintf(stderr, "dexpire: ending factor %6.2f\n", factor);

    return(0);
}

/*
 * calculate the free space and adjust the expiration munge
 * accordingly.
 */

int
NotEnoughFreeSpace(void)
{
    struct statfs stmp;
    int avail;
    char path[1024];

    sprintf(path, "%s/.", SpoolHome);

    if (FreeSpaceTarget == 0)
	return(0);
#if USE_SYSV_STATFS
    if (statfs(path, &stmp, sizeof(stmp), 0) != 0) {
#else
    if (statfs(path, &stmp) != 0) {
#endif
	fprintf(stderr, "dexpire: unable to statfs %s\n", path);
	return(0);
    }
    avail = stmp.f_bavail * (stmp.f_bsize / 512) / 2;	/* available space in KB */
    fprintf(stderr, "%4.2f MB free space availble\n", (double)avail / 1024.0);
    if (avail / 1024 < FreeSpaceTarget) {
	LotsOfFreeSpace = 0;
	return(1);
    }
    return(0);

#ifdef NOTDEF
	if (stmp.f_bavail < sfs.f_bavail)
	    sfs.f_bavail = stmp.f_bavail;
	if (stmp.f_ffree < sfs.f_ffree) {
	    sfs.f_ffree = stmp.f_ffree;
	    sfs.f_files = stmp.f_files;
	}
    }
#endif
}

void 
DoChildRemove(int pfd, int fd)
{
    long r;
    unsigned short nexp = (unsigned short)-1;

    while (read(pfd, &r, sizeof(long)) == sizeof(long)) {
       if (lseek(fd, r, 0) == r)
	   write(fd, &nexp, sizeof(nexp));
    }
}
